""" This module contains helper functions and data structures for interacting
with collectl and collectl generated data. More information on collectl can be
found at: http://collectl.sourceforge.net/.

"""
